import numpy as np

from .SimpfulModelBuilder import SugenoFISBuilder
from .Clustering import Clusterer
from .EstimateAntecendentSet import AntecedentEstimator
from .EstimateConsequentParameters import ConsequentEstimator
from .Tester import SugenoFISTester


x_train=np.array(range(105))

class ClusterChoser(object):
    def __init__(self, x_train, y_train, k_cross_val=10):
        self.x_train=x_train
        self.y_train=y_train
        
        def determineNumberClusters(self, x_train, y_train, k_cross_val=10, min_clusters=2, max_clusters=20, **kwargs):
            # shuffle data
            idx=np.arange(len(x_train))
            np.random.shuffle(idx)
            
            # split in k groups
            folds = np.array_split(idx,k_cross_val)
            
            # create an object of class Clusterer and set the required settings
            cl = Clusterer(self.x_train, self.y_train, nr_clus=0)
        
            if 'cluster_method' not in kwargs.keys(): kwargs['cluster_method'] = 'fcm'
        
            if kwargs['cluster_method'] == 'fcm':
                if 'fcm_m' not in kwargs.keys(): kwargs['fcm_m'] = 2
                if 'fcm_max_iter' not in kwargs.keys(): kwargs['fcm_maxiter'] = 1000
                if 'fcm_error' not in kwargs.keys(): kwargs['fcm_error'] = 0.005
    
            elif kwargs['cluster_method'] == 'fstpso':
                if 'fstpso_n_particles' not in kwargs.keys(): kwargs['fstpso_n_particles'] = None
                if 'fstpso_maxiter' not in kwargs.keys(): kwargs['fstpso_maxiter'] = 100
                if 'fstpso_path_fit_dump' not in kwargs.keys(): kwargs['fstpso_path_fit_dump'] = None
                if 'fstpso_path_sol_dump' not in kwargs.keys(): kwargs['fstpso_path_sol_dump'] = None
                
            if 'mf_shape' not in kwargs.keys(): kwargs['mf_shape'] = 'gauss'
            if 'global_fit' not in kwargs.keys(): kwargs['global_fit'] = True 
            if 'save_simpful_code' not in kwargs.keys(): kwargs['save_simpful_code'] = False
            if 'operators' not in kwargs.keys(): kwargs['operators'] = None
            
            for c in range (min_clusters,max_clusters):
                cl.nr_clus=c
                for k in range(0,k_cross_val): 
                    # select the indices used for validation
                    idx_val = folds[k]
                    
                    # Use other indices for training
                    idx_train = list(set(range(len(idx))) - set(idx_val))
                    
                    # Select training data for this interation
                    trx=np.array(idx_train)
                    cl.x_train=x_train[trx]
                    cl.y_train=x_train[trx]

                    # Cluster the training data (in input-output space) using FCM or FST-PSO
                    if kwargs['cluster_method'] == 'fcm':
                        self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
                                                                                    fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
                    elif kwargs['cluster_method'] == 'fstpso':
                        self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
                                                                                    fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                                                                                    fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
                    
                    # Estimate the membership funtions of the system (default shape: gauss)
                    self._antecedent_estimator = AntecedentEstimator(self.x_train, self.partition_matrix)
                    self.antecedent_parameters = self._antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=merge_threshold)
                    what_to_drop = self._antecedent_estimator._info_for_simplification
        
                    # Estimate the parameters of the consequent (default: global fitting)
                    ce = ConsequentEstimator(self.x_train, self.y_train, self.partition_matrix)
                    self.consequent_parameters = ce.suglms(self.x_train, self.y_train, self.partition_matrix, 
                                               global_fit=kwargs['global_fit'])
        
                    # Build a first-order Takagi-Sugeno model using Simpful
                    simpbuilder = SugenoFISBuilder(
                            self.antecedent_parameters, 
                            self.consequent_parameters, 
                            self.variable_names, 
                            extreme_values = self._antecedent_estimator._extreme_values,
                            operators=kwargs["operators"], 
                            save_simpful_code=kwargs['save_simpful_code'], 
                            fuzzy_sets_to_drop=what_to_drop)

                    self.model = simpbuilder.simpfulmodel

            
            
            
            


from .LoadData import DataLoader
from .Splitter import DataSplitter
from .SimpfulModelBuilder import SugenoFISBuilder
from .Clustering import Clusterer
from .EstimateAntecendentSet import AntecedentEstimator
from .EstimateConsequentParameters import ConsequentEstimator
from .Tester import SugenoFISTester
import numpy as np


class FeatureSelector(object):
    def __init__(self, datapath, nr_clus, variable_names,feature_selection_threshold=0.05, **kwargs):
        self.dataX
        self.dataY
        self.nr_clus = nr_clus
        self.variable_names = variable_names
       
        # Check settings and complete with defaukt settings when needed
        if 'normalize' not in kwargs.keys(): kwargs['normalize'] = False
        if 'merge_threshold' not in kwargs.keys(): kwargs['merge_threshold'] = 1.0
        if 'cluster_method' not in kwargs.keys(): kwargs['cluster_method'] = 'fcm'        
        if kwargs['cluster_method'] == 'fcm':
            if 'fcm_m' not in kwargs.keys(): kwargs['fcm_m'] = 2
            if 'fcm_max_iter' not in kwargs.keys(): kwargs['fcm_maxiter'] = 1000
            if 'fcm_error' not in kwargs.keys(): kwargs['fcm_error'] = 0.005
            self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
                fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
        elif kwargs['cluster_method'] == 'fstpso':
            if 'fstpso_n_particles' not in kwargs.keys(): kwargs['fstpso_n_particles'] = None
            if 'fstpso_maxiter' not in kwargs.keys(): kwargs['fstpso_maxiter'] = 100
            if 'fstpso_path_fit_dump' not in kwargs.keys(): kwargs['fstpso_path_fit_dump'] = None
            if 'fstpso_path_sol_dump' not in kwargs.keys(): kwargs['fstpso_path_sol_dump'] = None
            self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
                fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
        if 'mf_shape' not in kwargs.keys(): kwargs['mf_shape'] = 'gauss'       
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
        if 'global_fit' not in kwargs.keys(): kwargs['global_fit'] = False  
        if 'save_simpful_code' not in kwargs.keys(): kwargs['save_simpful_code'] = True
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
                            

        # Create a training and valiadation set for the feature selection phase
        ds = DataSplitter(self.dataX, self.dataY)
        x_feat, y_feat, x_val, y_val = ds.holdout()
        
        # Set initial values for the MAEs
        old_MAE=-1
        new_MAE=feature_selection_threshold
        MAEs=[]
        selected_features=[]
        stop=False

        while stop == False:   
            for f in range[0,np.size(x_feat,axis=1)]:
                var_names=selected_features
                feat=np.concatenate((x_feat[:,selected_features], x_feat[:,f]),axis=1)
                
                # Cluster the training data (in input-output space)
                cl = Clusterer(feat, y_feat, self.nr_clus)               
                
                if kwargs['cluster_method'] == 'fcm':
                    cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
                        fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
                elif kwargs['cluster_method'] == 'fstpso':
                    cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
                        fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                        fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
                else:
                    print('The requested clustering method is not (yet) implemented')
                     
                # Estimate the membership funtions of the system (default shape: gauss)
                antecedent_estimator = AntecedentEstimator(feat, partition_matrix)
        
                antecedent_parameters = antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=merge_threshold)
                what_to_drop = antecedent_estimator._info_for_simplification
        
                # Build a first-order Takagi-Sugeno model using Simpful using dummy consequent parameters
                simpbuilder = SugenoFISBuilder(
                    antecedent_parameters, 
                    np.tile(1, (self.nr_clus, len(self.variable_names)+1)), 
                    self.variable_names, 
                    extreme_values = antecedent_estimator._extreme_values,
                    operators=kwargs["operators"], 
                    save_simpful_code=False, 
                    fuzzy_sets_to_drop=what_to_drop)
        
                dummymodel = simpbuilder.simpfulmodel
                
                # Calculate the firing strengths for each rule for each data point 
                firing_strengths=[]
                for i in range(0,len(self.x_train)):
                    for j in range (0,len(var_names)):
                        dummymodel.set_variable(var_names[j], feat[i,j])
                    firing_strengths.append(dummymodel.get_firing_strengths())
                firing_strengths=np.array(firing_strengths)
                
                
                # Estimate the parameters of the consequent
                ce = ConsequentEstimator(feat, y_feat, firing_strengths)
                consequent_parameters = ce.suglms(feat, y_feat, firing_strengths, 
                                                       global_fit=kwargs['global_fit'])
        
                # Build a first-order Takagi-Sugeno model using Simpful
                simpbuilder = SugenoFISBuilder(
                    antecedent_parameters, 
                    consequent_parameters, 
                    variable_names, 
                    extreme_values = antecedent_estimator._extreme_values,
                    operators=kwargs["operators"], 
                    save_simpful_code=kwargs['save_simpful_code'], 
                    fuzzy_sets_to_drop=what_to_drop)
        
                model = simpbuilder.simpfulmodel
                
                MAEs[f] = model.calculate_error(method="MAE")
            
            new_MAE=max(MAEs)
            new_feature=variable_names[MAEs.index(new_MAE)]
            del MAEs
            
            if new_MAE-old_MAE>old_MAE*feature_selection_threshold:
                selected_features.append(new_feature)
                old_MAE=new_MAE
            else:
                print('The following features were selected:',  selected_features)
                stop = True 
                    
                    



import numpy as np
from scipy.optimize import curve_fit
from sklearn.linear_model import LinearRegression
import numpy.matlib

class ConsequentEstimator(object):
    def __init__(self, x_train, y_train, partition_matrix):
        self.x_train=x_train
        self.y_train=y_train
        self.partition_matrix=partition_matrix
        
    def suglms(self, x_train, y_train, partition_matrix, global_fit=True, df=0):
        # SUGLMS estimates the consequent parameters in the Sugeno-Takagi model
        #     using least squares.
        #
        #    Input:
        #       X .....    input data matrix
        #        Y .....    output data vector
        #       F ..... fuzzy partition matrix (membership degrees),
        #                optional, defaults to ones(size(y)) for
        #                which SUGLMS is a standard linear regression
        #       DF ... default value returned when the sum of grades
        #               equals to one (optional, defaults to 0)
        #        FLAG .. set to 1 to get local weighted LMS estimates
        #    
        #    Output:
        #       P .....    consequents parameters for every cluster
        #        Ym ....    global model output for the given input data
        #        Yl ....    output of local submodels (corresponding to clusters)
        #        Ylm ...    output of local submodels with data corresponding to
        #               degrees < 0.2 masked with NaN's (for plots)
        #
        #    Example:
        #    x = (0:0.02:1)'; y = sin(7*x);
        #       f = mgrade(x',mfequ(x,2,3))';
        #       [p,ym,yl,ylm] = suglms([x ones(size(x))],y,f);
        #    subplot(211); plot(x,ylm,'.',x,[y ym]); title('Fitting y = sin(7*x)')
        #    subplot(212); plot(x,f); title('Membership functions')
        # (c) Robert Babuska, 1994-95
        
#        x_train=[11,12,13
#                 21,22,23
#                 31,32,33
#                 41,42,43
#                 51,52,53]
#        
#        y_train=[1
#                 2
#                 3
#                 4
#                 5]
#        
#        partition_matrix=[0.2, 0.8
#                          0.4, 0.6
#                          0.6, 0.4
#                          0.8, 0.2
#                          1.0, 0.0]
#        
#        
        x=x_train.copy()
        y=y_train.copy()
        f=partition_matrix.copy()
        
        # Check if input X contains one column of ones (for the constant). If not, add it.
        u=np.unique(x[:,-1])
        if u.shape[0]!=1 or u[0]!=1:
            x = np.hstack((x,np.ones((x.shape[0],1))))
    
        # Find the number of data points (mx & mx) , the number of variables (nx) and the
        # number of clusters (nf) 
        mx,nx=x.shape
        mf,nf=f.shape
        
        # Calculate the sum of the degree of fulfillement (DOF) for each data point
        sumDOF=np.sum(f, 1)
        
        
        # When degree of fulfillment is zero (which means no rule is applicable), set to one
        NoRule = sumDOF == 0
        sumDOF[NoRule] = 1
        sumDOF = np.matlib.repmat(sumDOF,nf,1).T
        
        # Auxillary variables
        f1 = x.flatten()
        s = np.matlib.repmat(f1,nf,1).T
        xx = np.reshape(s, (nx,nf*mx), order='F')
        s = xx.T  
        x1=np.reshape(s,(mx,nf*nx),order='F') 
        x=x.T                                # reshape data matrix
        
        
        if nf == 1:
           global_fit = False
        
        if global_fit == True:                                          # Global least mean squares estimates 
            # (reshaped) vector of f devided by the sum of each row of f
            # (normalised membership degree)
            xx = (f.T.flatten()/sumDOF.T.flatten())
            
            # reshape partition matrix
            s= np.matlib.repmat(xx,nx,1).T
            f1 = np.reshape(s, (mx,nf*nx), order='F')                
            x1 = f1*x1
            y1 = 
        
            # Find least squares solution
#            xp = np.linalg.lstsq(x1,y,rcond=None)
            
            # Perform QR decomposition
            Q,R = np.linalg.qr(x1)          # qr decomposition of x1
            Qy = np.dot(Q.T,y1)             # computing Q^T*b (project y onto the range of x1)
            xx = np.linalg.solve(R,Qy)
            
            
        elif global_fit == False:           # local weighted least mean squares estimates
            # Pre-allocate variable
            p=np.zeros([nf,nx])
            for i in range (0,nf):
                # Weight the points with the cluster membership
                W=np.sqrt(np.diag(f[:,i]))
                x1 = np.dot(W,x)
                y1 = np.dot(W,y)
               
                # Perform QR decomposition
                Q,R = np.linalg.qr(x1)      # qr decomposition of x1
                Qy = np.dot(Q.T,y1)         # computing Q^T*b (project by onto the range of x1)
                
                #perform least-squares
                p[i] = np.linalg.solve(R,Qy)
        return p
    
    def sugfunc(self, x1, x2, a, b, c):
        return a*x1 + b*x2 + c   
    


import numpy as np

class DataLoader(object):
        def __init__(self, datapath, variable_names=None, normalize=False):
            if variable_names is None:
                variable_names = np.genfromtxt(datapath, dtype='str', delimiter=',', max_rows=1)
                print('The following variable names were detected from the input file: ', variable_names)
                self.output_name=variable_names[-1]
                self.variable_names=variable_names[:-1]
                self.data=np.loadtxt(datapath, delimiter=',', skiprows=1)
            else:
                self.variable_names=variable_names
                self.data=np.loadtxt(datapath, delimiter=',')
                
            self.dataX=self.data[:,0:-1]
            self.dataY=self.data[:,-1]
            
            if normalize=='minmax' or normalize=='linear' or normalize==True:
                self.dataX = (self.dataX - np.abs(self.dataX).min(axis=0)) / (np.abs(self.dataX).max(axis=0)-np.abs(self.dataX).min(axis=0))

            elif normalize == 'zscore':
                self.dataX = (self.dataX - self.dataX.mean(axis=0)) / self.dataX.std(axis=0)
            



import numpy as np
from scipy.spatial.distance import cdist
from random import seed

# To do:
# Give option to cluster in input or input-output space?
# More clustering methods

class Clusterer(object):
    def __init__(self, x_train, y_train, nr_clus):
        self.x_train=x_train
        self.y_train=y_train
        self.data=np.concatenate((self.x_train,self.y_train.reshape(len(self.y_train),1)),axis=1)
        self.nr_clus = nr_clus
        

    def cluster(self, method="fcm", **kwargs):
        try:
            m = kwargs["m"]
        except:
            m = 2

        if method == "fcm":
            try:
                max_iter = kwargs["fcm_max_iter"]
            except:
                max_iter = 1000
            try:
                error = kwargs["fcm_error"]
            except:
                error = 0.005
#            try:
#                seed = kwargs["seed"]
#            except:
#                seed = None
            centers, partition_matrix, jm = self._fcm(data=self.data, n_clusters=self.nr_clus, m=m, max_iter=max_iter, error=error)
        
        elif method== "pfcm":
            try:
                n = kwargs["pfcm_n"]
            except:
                n = 2
            try:
                max_iter = kwargs["pfcm_max_iter"]
            except:
                max_iter = 1000
            try:
                error = kwargs["pfcm_error"]
            except:
                error = 0.005
            try:
                a = kwargs["pfcm_a"]
            except:
                a = 0.5
            try:
                b = kwargs["pfcm_b"]
            except:
                b = 0.5

            centers, partition_matrix, typicality_matrix, jm = self._pfcm(data=self.data, n_clusters=self.nr_clus, m=m, n=n, max_iter=max_iter, error=error, a=a, b=b)    

        elif method == "fstpso":
            try:
                max_iter = kwargs["fstpso_max_iter"]
            except:
                max_iter=100
            try:
                n_particles = kwargs["fstpso_n_particles"]
            except:
                n_particles = None
            try:
                path_fit_dump = kwargs["fstpso_path_fit_dump"]
            except:
                path_fit_dump = None
            try:
                path_sol_dump = kwargs["fstpso_path_sol_dump"]
            except:
                path_sol_dump = None

            centers, partition_matrix, jm = self._fstpso(data=self.data, n_clusters=self.nr_clus, max_iter=max_iter, n_particles=n_particles, m=m, path_fit_dump=path_fit_dump, path_sol_dump=path_sol_dump)

        return centers, partition_matrix, jm

    def _fcm(self, data, n_clusters, m=2, max_iter=1000, error=0.005, seed=None):
        #data: 2d array, size (N, S). N is the number of instances; S is the number of variables
        #n_clusters: number of clusters
        #m: fuzzy clustering coefficient
        #max_it: maximum number of iterations, default=1000
        #error: stopping criterion, default=0.005
        #seed: seed for random initialization of u matrix
        
        n_instances = data.shape[0]
        
        #randomly initaliaze u
        if seed != None:
            np.random.seed(seed=seed)
        u = np.random.rand(n_instances, n_clusters)
        u = np.fmax(u, np.finfo(np.float64).eps)
        ut = u.T
        
        for it in range(0,max_iter):
            #copy old u matrix
            u_old = ut.copy()
            u_old /= np.ones((n_clusters, 1)).dot(np.atleast_2d(u_old.sum(axis=0)))
            u_old = np.fmax(u_old, np.finfo(np.float64).eps)
        
            #elevate to m
            um = u_old ** m
        
            #calculate cluster centers
            centers = um.dot(data) / (np.ones((data.shape[1], 1)).dot(np.atleast_2d(um.sum(axis=1))).T)
        
            #calculate distances
            dist = cdist(centers, data, metric='euclidean')
            dist = np.fmax(dist, np.finfo(np.float64).eps)
        
            #calculate objective
            jm = (um * dist ** 2).sum()
        
            #calculate new u matrix
            ut = dist ** (- 2. / (m - 1))
            ut /= np.ones((n_clusters, 1)).dot(np.atleast_2d(ut.sum(axis=0)))
        
            #stopping criterion
            if np.linalg.norm(ut - u_old) < error:
                break
        
        partition_matrix = ut.T
        return centers, partition_matrix, jm
    
    
    def _fstpso(self, data, n_clusters, max_iter=100, n_particles=None, m=2, path_fit_dump=None, path_sol_dump=None):
        #data: 2d array, size (N, S). N is the number of instances; S is the number of variables
        #n_clusters: number of clusters
        #max_iter: number of maximum iterations of FST-PSO, default is 100
        #n_particles: number of particles in the swarm, if None it is automatically set by FST-PSO
        #m: fuzzy clustering coefficient
        #path_fit_dump: path to the file where the best fitness score at each iteration will be dumped
        #path_sol_dump: path to the file where the best solution at each iteration will be dumped

        try:
            from fstpso import FuzzyPSO
        except:
            print("ERROR: please, pip install fst-pso to use this functionality.")

        n_instances = data.shape[0]
        n_variables = data.shape[1]

        #set search space boundaries
        bounds = [0]*n_variables
        for i in range(n_variables):
            x = min([row[i] for row in data])
            y = max([row[i] for row in data])
            bounds[i] = [x, y]

        search_space = []
        for i in bounds:
            search_space.extend([i]*n_clusters)

        #initializing FST-PSO
        FP = FuzzyPSO()
        FP.set_search_space(search_space)
        if n_particles != None: FP.set_swarm_size(n_particles)

        #generally better results are obtained with this rule disabled
        FP.disable_fuzzyrule_minvelocity()

        #fitness function definition
        def fitness(particle):
            particle = list(map(float,particle))
            centers = np.reshape(particle, (n_variables, n_clusters)).T

            #calculating fitness value of found solution
            dist = cdist(data, centers, metric='sqeuclidean')
            
            um = np.zeros(np.shape(dist))
            for i in range(np.shape(um)[0]):
                for j in range(np.shape(um)[1]):
                    um[i][j] = np.sum(    np.power(    np.divide(    dist[i][j],dist[i])    ,    float(1/(m-1))    )    )
            um = np.reciprocal(um)

            um_power = np.power(um,m)

            fitness_value = np.sum(np.multiply(um_power,dist))
            return fitness_value

        #fitness function setting
        FP.set_fitness(fitness, skip_test=True)

        #execute optimization
        result = FP.solve_with_fstpso(max_iter=max_iter, dump_best_fitness=path_fit_dump, dump_best_solution=path_sol_dump)

        #reshaping centers
        solution = list(map(float,result[0].X))
        centers = np.reshape(solution, (n_variables, n_clusters)).T
        
        #calculating membership matrix
        dist = cdist(data, centers, metric='sqeuclidean')
        um = np.zeros(np.shape(dist))
        for i in range(np.shape(um)[0]):
            for j in range(np.shape(um)[1]):
                um[i][j] = np.sum(    np.power(    np.divide(    dist[i][j],dist[i])    ,    float(1/(m-1))    )    )
        partition_matrix = np.reciprocal(um)

        #final fitness value
        jm =  result[1]

        return centers, partition_matrix, jm
    
    def _pfcm(self, data, n_clusters, m=2, max_iter=1000, error=0.005, a=0.5, b=0.5, n=2, seed=None):
        """
        data: Dataset to be clustered, with size M-by-N, where M is the number of data points
        and N is the number of coordinates for each data point.
        c : Number of clusters
        m: fuzzy clustering coefficient (default = 2)
        max_iter: Maximum number of iterations (default = 1000)
        error : stopping criterion (default=0.005)
        a: Relative importane of fuzzy membership (default = 0.5)
        b: Relative importane of typicality values (default = 0.5) 
        n: User-defined constant n (default = 2)

        Return values:
        centers: The locations of the found clusters centers
        partition_matrix: Partition matrix
        typicality_matrix: Typicality Matrix
        jm: The objective funtion for U and T
        """

        # Randomly initiaize the partitioning matrix and typicality matrix
        if seed != None:
            np.random.seed(seed=seed)
        u = np.random.rand(n_instanes, n_clusters)
        u = np.fmax(u, np.finfo(np.float64).eps)        # avoid 0's in the matrix
        t = np.random.rand(n_instanes, n_clusters)
        t = np.fmax(t, np.finfo(np.float64).eps)        # avoid 0's in the matrix

        # Pre-allocation
        jm = np.zeros(shape=(max_iter, 1))
        g = np.zeros(shape=(c, data.shape[0]))


        for i in range(max_iter):
           
            # Perform one iteration of PFCM
            partition_matrix, typicality_matrix, centers, jm[i], g = pstepfcm(data, cntr, U, T, m, a, b, n, g)
            
            # Stopping criterion: Stop if objective value does inrease < error
            if abs(jm[i] - jm[i-1]) < error:
                break

        return centers, partition_matrix, typicality_matrix, jm


    def _pstepfcm(self, data, centers, U, T, g, m=2, n=2, a=0.5, b=0.5):
        #copy old u and t matrix and center locations
        um = U**m
        tf = T**n
        tfo = (1-T) ** n
        centers = (np.dot(a*um+b*tf, data).T/np.sum(a*um+b*tf, axis=1).T).T

        # Calculate distances between data points and cluster centers            
        dist = cdist(centers, data, metric='euclidean')
        dist = np.fmax(dist, np.finfo(np.float64).eps)

        #calculate value of objective funtion
        jm = np.sum(np.sum(np.power(dist, 2)*(a*um+b*tf), axis=0)) + np.sum(g*np.sum(tfo, axis=0))

        #calculate new u and t matrix
        g = um*np.power(dist, 2)/(np.sum(um, axis=0))
        tmp = np.power(dist, (-2/(m-1)))
        U = tmp/(np.sum(tmp, axis=0))
        tmpt = np.power((b/g)*np.power(dist, 2), (1/(n-1)))
        T = 1/(1+tmpt)

        return U, T, centers, jm, g




import numpy as np
from scipy.optimize import curve_fit
from sklearn.linear_model import LinearRegression
import numpy.matlib
import statsmodels.api as sm

class ConsequentEstimator(object):
    def __init__(self, x_train, y_train, firing_strengths):
        self.x_train=x_train
        self.y_train=y_train
        self.firing_strengths=firing_strengths
        
    def suglms(self, x_train, y_train, firing_strengths, global_fit=False, df=0):
        # SUGLMS estimates the consequent parameters in the Sugeno-Takagi model
        #	 using least squares.
        #
        #    Input:
        #       X .....	input data matrix
        #	    Y .....	output data vector
        #       F ..... fuzzy partition matrix (membership degrees),
        #		        optional, defaults to ones(size(y)) for
        #		        which SUGLMS is a standard linear regression
        #       DF ... default value returned when the sum of grades
        #               equals to one (optional, defaults to 0)
        #	    FLAG .. set to 1 to get local weighted LMS estimates
        #    
        #    Output:
        #       P .....	consequents parameters for every cluster
        #	    Ym ....	global model output for the given input data
        #	    Yl ....	output of local submodels (corresponding to clusters)
        #	    Ylm ...	output of local submodels with data corresponding to
        #               degrees < 0.2 masked with NaN's (for plots)
        #
        #    Example:
        #	x = (0:0.02:1)'; y = sin(7*x);
        #       f = mgrade(x',mfequ(x,2,3))';
        #       [p,ym,yl,ylm] = suglms([x ones(size(x))],y,f);
        #	subplot(211); plot(x,ylm,'.',x,[y ym]); title('Fitting y = sin(7*x)')
        #	subplot(212); plot(x,f); title('Membership functions')
        # (c) Robert Babuska, 1994-95
        
        x=x_train.copy()
        y=y_train.copy()
        f=firing_strengths.copy()
        
        # Check if input X contains one column of ones (for the constant). If not, add it.
        u=np.unique(x[:,-1])
        if u.shape[0]!=1 or u[0]!=1:
            x = np.hstack((x,np.ones((x.shape[0],1))))
    
        # Find the number of data points (mx & mf) , the number of variables (nx) and the
        # number of clusters (nf) 
        mx,nx=x.shape
        mf,nf=f.shape
        
        # Calculate the sum of the degree of fulfillement (DOF) for each data point
        sumDOF=np.sum(f, 1)
        
        
        # When degree of fulfillment is zero (which means no rule is applicable), set to one
        NoRule = sumDOF == 0
        sumDOF[NoRule] = 1
        sumDOF = np.matlib.repmat(sumDOF,nf,1).T
        
        
        if nf == 1:
           global_fit = False
        
        if global_fit == True:                # Global least mean squares estimates 

            # Still under construction!
            
            # Auxillary variables
            f1 = x.flatten()
            s = np.matlib.repmat(f1,nf,1).T
            xx = np.reshape(s, (nx,nf*mx), order='F')
            s = xx.T  
            x1=np.reshape(s,(mx,nf*nx),order='F') 
            x=x.T                                # reshape data matrix
            
            # (reshaped) vector of f devided by the sum of each row of f
            # (normalised membership degree)
            xx = (f.T.flatten()/sumDOF.T.flatten())
            
            # reshape partition matrix
            s= np.matlib.repmat(xx,nx,1).T
            f1 = np.reshape(s, (mx,nf*nx), order='F')                
            x1 = f1*x1
        
            # Find least squares solution
#            xp = np.linalg.lstsq(x1,y,rcond=None)
            
            # Perform QR decomposition
            Q,R = np.linalg.qr(x1) # qr decomposition of A
            Qy = np.dot(Q.T,y) # computing Q^T*b (project b onto the range of A)
            xx = np.linalg.solve(R,Qy)
            
            p=np.reshape(xx,(nf,nx), order='F')
        
            # Local models
            yl = np.transpose(x).dot(np.transpose(p))						                    #
            
            # Global model
            ym = x1.dot(p.flatten()) + df*NoRule
            ylm = yl.copy()
            
            # Mask all memberships < 0.2 with NaN's for plots
            ylm[f<0.2] = np.NaN
        
        elif global_fit == False:                                         # local weighted least mean squares estimates
            # Pre-allocate variable
            p=np.zeros([nf,nx])
            
            for i in range (0,nf):
                '''
                # Weight the points with the firing strength of the rule
                #W=np.sqrt(np.diag(f[:,i]))
                W=np.array([f[:,i],]*nx).transpose()
                x1 = np.multiply(W,x)
                y1 = np.multiply(f[:,i],y)
               
                # Perform QR decomposition
                Q,R = np.linalg.qr(x1)      # qr decomposition of x1
                Qy = np.dot(Q.T,y1)         # computing Q^T*b (project by onto the range of x1)
                
                #perform least-squares
                p[i] = np.linalg.solve(R,Qy)
                
                wls_model=sm.WLS(y, x, weights=f[:,i]) 
                results = wls_model.fit()
                p[i]=np.array(results.params)               
        '''
                # Select firing strength of the selected rule
                w= f[:,i]
                
                # Weight input with firing strength
                xw = x * np.sqrt(w[:,np.newaxis])
                
                # Weight output with firing strength
                yw = y * np.sqrt(w)
                
                # Perform least squares with weighted input and output
                prm,_,_,_=np.linalg.lstsq(xw, yw, rcond=None)
                p[i]=prm
                       
        return p #,ym,yl,ylm
    
    def sugfunc(self, x1, x2, a, b, c):
        return a*x1 + b*x2 + c   
    


from .Splitter import DataSplitter
from .SimpfulModelBuilder import SugenoFISBuilder
from .Clustering import Clusterer
from .EstimateAntecendentSet import AntecedentEstimator
from .EstimateConsequentParameters import ConsequentEstimator
from .Tester import SugenoFISTester
import numpy as np

class FeatureSelector(object):
    def __init__(self, dataX, dataY, nr_clus, variable_names, **kwargs):
        self.dataX=dataX
        self.dataY=dataY
        self.nr_clus = nr_clus
        self.variable_names = variable_names
       

                            
    def wrapper(self, feature_selection_stop=0.005,**kwargs):
        
        # Check settings and complete with defaukt settings when needed
        if 'merge_threshold' not in kwargs.keys(): kwargs['merge_threshold'] = 1.0
        if 'cluster_method' not in kwargs.keys(): kwargs['cluster_method'] = 'fcm'        
        if kwargs['cluster_method'] == 'fcm':
            if 'fcm_m' not in kwargs.keys(): kwargs['fcm_m'] = 2
            if 'fcm_max_iter' not in kwargs.keys(): kwargs['fcm_maxiter'] = 1000
            if 'fcm_error' not in kwargs.keys(): kwargs['fcm_error'] = 0.005
        elif kwargs['cluster_method'] == 'fstpso':
            if 'fstpso_n_particles' not in kwargs.keys(): kwargs['fstpso_n_particles'] = None
            if 'fstpso_max_iter' not in kwargs.keys(): kwargs['fstpso_max_iter'] = 100
            if 'fstpso_path_fit_dump' not in kwargs.keys(): kwargs['fstpso_path_fit_dump'] = None
            if 'fstpso_path_sol_dump' not in kwargs.keys(): kwargs['fstpso_path_sol_dump'] = None
        if 'mf_shape' not in kwargs.keys(): kwargs['mf_shape'] = 'gauss'       
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
        if 'global_fit' not in kwargs.keys(): kwargs['global_fit'] = False  
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
        
        
        # Create a training and valiadation set for the feature selection phase
        ds = DataSplitter(self.dataX, self.dataY)
        x_feat, y_feat, x_val, y_val = ds.holdout(self.dataX, self.dataY)
        
        # Set initial values for the MAEs
        old_MAE=np.inf
        new_MAE=np.inf
        MAEs=[]
        
        # Create a set with the unselected (currently all) and selected (none yet) variables
        selected_features=[]
        unselected_features=list(range(0,np.size(x_feat,axis=1)))
        
        stop=False

        while stop == False: 
            MAEs= [np.inf]*np.size(x_feat,axis=1)
            
            for f in [x for x in unselected_features if x != -1]:
                considered_features = selected_features + [f]
                var_names=self.variable_names[considered_features] 
                feat=x_feat[:,considered_features]

                
                # Cluster the training data (in input-output space)
                cl = Clusterer(feat, y_feat, self.nr_clus)               
                
                if kwargs['cluster_method'] == 'fcm':
                    cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
                        fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
                elif kwargs['cluster_method'] == 'fstpso':
                    cluster_centers, partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
                        fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                        fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
                else:
                    print('The requested clustering method is not (yet) implemented')
                     
                # Estimate the membership funtions of the system (default shape: gauss)
                antecedent_estimator = AntecedentEstimator(feat, partition_matrix)
        
                antecedent_parameters = antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=kwargs['merge_threshold'])
                what_to_drop = antecedent_estimator._info_for_simplification
        
                # Build a first-order Takagi-Sugeno model using Simpful using dummy consequent parameters
                simpbuilder = SugenoFISBuilder(
                    antecedent_parameters, 
                    np.tile(1, (self.nr_clus, len(var_names)+1)), 
                    var_names, 
                    extreme_values = antecedent_estimator._extreme_values,
                    operators=kwargs["operators"], 
                    save_simpful_code=False, 
                    fuzzy_sets_to_drop=what_to_drop)
        
                dummymodel = simpbuilder.simpfulmodel
                
                # Calculate the firing strengths for each rule for each data point 
                firing_strengths=[]
                for i in range(0,len(feat)):
                    for j in range (0,len(var_names)):
                        dummymodel.set_variable(var_names[j], feat[i,j])
                    firing_strengths.append(dummymodel.get_firing_strengths())
                firing_strengths=np.array(firing_strengths)
                
                
                # Estimate the parameters of the consequent
                ce = ConsequentEstimator(feat, y_feat, firing_strengths)
                consequent_parameters = ce.suglms(feat, y_feat, firing_strengths, 
                                                       global_fit=kwargs['global_fit'])
        
                # Build a first-order Takagi-Sugeno model using Simpful
                simpbuilder = SugenoFISBuilder(
                    antecedent_parameters, 
                    consequent_parameters, 
                    var_names, 
                    extreme_values = antecedent_estimator._extreme_values,
                    operators=kwargs["operators"], 
                    save_simpful_code=False, 
                    fuzzy_sets_to_drop=what_to_drop)
        
                model = simpbuilder.simpfulmodel
                
                # Test the model
                test = SugenoFISTester(model, x_val[:,considered_features],y_val)
                mae = test.calculate_MAE(variable_names=var_names)
                
                MAEs[f] = mae
            
            new_MAE=min(MAEs)
            #print('Unselected feaures:', unselected_features)
            #print('index new MAE:', MAEs.index(new_MAE))
            new_feature=unselected_features[MAEs.index(new_MAE)]
            
            #print('new feature', new_feature)
            del MAEs
            
            if new_MAE<old_MAE: #*feature_selection_stop:
                selected_features.append(new_feature)
                unselected_features[new_feature]=-1
                old_MAE=new_MAE
            else:
                
                stop = True 
       
        selected_feature_names = self.variable_names[selected_features]
        print('The following features were selected:',  selected_feature_names)
        
        return selected_features, selected_feature_names       
    
   



import simpful 

class SimpfulConverter(object):
    """    This object converts a description of a Fuzzy System into a readable
        Simpful project file.
    """
    
    def __init__(self, 
        input_variables_names,
        consequents_matrix,
        fuzzy_sets,
        fuzzy_sets_to_drop=dict,
        extreme_values = None,
        operators = None
        ):
        super(SimpfulConverter, self).__init__()
        self._input_variables = input_variables_names
        self._consequents_matrix = consequents_matrix
        self._clusters = len(self._consequents_matrix)
        self._fuzzy_sets = fuzzy_sets
        self._fuzzy_sets_to_drop = fuzzy_sets_to_drop
        self._extreme_values = extreme_values

        assert(len(self._input_variables)+1 == len(self._consequents_matrix[0]))

        print(" * Detected %d rules / clusters" % self._clusters)

        self._source_code = []
        self._source_code.append( '# WARNING: this source code was automatically generated by pyFUME.' )
        self._source_code.append( "from simpful import *" )
        if operators is None:
            self._source_code.append("\nFS = FuzzySystem()")
        else:
            # experimental, please test ASAP
            self._source_code.append("\nFS = FuzzySystem(operators="+str(operators)+")")

        
    def save_code(self, path):
        code = self.generate_code()
        with open(path, "w") as fo:
            fo.write(code)
        print (" * Code saved to file %s" % path)

    def generate_object(self):
        code = self.generate_code()
        exec(code, globals()) 
        from copy import deepcopy
        self._fuzzyreasoner = deepcopy(FS)

    def generate_code(self, use_main=False):

        # rules
        rule_texts = self.create_rules()
        for i in range(1, self._clusters+1):
            self._source_code.append('RULE%d = "%s"' % (i, rule_texts[i-1]))
        self._source_code.append("FS.add_rules([%s])" % (", ". join(["RULE%d" % i for i in range(1,self._clusters+1)])))

        self._source_code.append("")

        # output functions
        B = self._create_consequents()
        for i in range(self._clusters):
            self._source_code.append("FS.set_output_function('%s', '%s')" % ("fun%d" % (i+1),  B[i]))

        self._source_code.append("")
    
        # fuzzy sets and membership functions
        result = self._create_fuzzy_sets()
        self._source_code.append(result)

        self._source_code.append("# end of automatically generated code #")

        return "\n".join(self._source_code)


    def _create_fuzzy_sets(self, verbose=False):
        j=0
        chunk = ""
        for num_var, var in enumerate(self._input_variables):

            subchunk = []
            for cluster in range(self._clusters):

                if (num_var, cluster) in self._fuzzy_sets_to_drop:
                    chunk+="# "

                if verbose: print (" * Creating fuzzy set for variable %s, cluster%d" % (var, cluster+1))
                chunk += 'FS_%d = FuzzySet(' % (j+1)

                term = 'cluster%d' % (cluster+1)

                fstype, params = self._fuzzy_sets[j]
                if fstype == 'gauss':
                    chunk += "function=Gaussian_MF(%f, %f), term='%s')" % (params[0], params[1], term) 
 
                elif fstype == 'gauss2':
                    chunk += "function=DoubleGaussian_MF(%f, %f, %f, %f), term='%s')" % (params[0], params[1], params[2], params[3], term) 

                elif fstype == 'sigmoid':
                    chunk += "function=Sigmoid_MF(%f, %f), term='%s')" % (params[0], params[1], term) 

                elif fstype == 'invgauss':
                    chunk += "function=InvGaussian_MF(%f, %f), term='%s')" % (params[0], params[1], term) 
                else:
                    raise Exception("Fuzzy set type not supported,"+fstype)
                if (num_var, cluster) not in self._fuzzy_sets_to_drop:
                    subchunk.append("FS_%d" % (j+1))
                #print ( self._fuzzy_sets[j] )
                j += 1
                chunk += "\n"

            chunk += "MF_%s = LinguisticVariable([%s], concept='%s' , universe_of_discourse=%s)\n" % (var, ", ".join(subchunk), var, self._extreme_values[num_var] )
            chunk += "FS.add_linguistic_variable('%s', MF_%s)\n\n" % (var, var)

        return chunk

    def _create_consequents(self):
        result = []
        for row in self._consequents_matrix:
            result.append(("+".join(["%.2e*%s" % (value, name) for (name, value) in zip(self._input_variables, row[:-1])])))
            result[-1] += "+%.2e" % row[-1]
        return result

    def _create_antecedents(self):

        result = []

        for i in range(self._clusters):
            
            pieces = []
            for j, var in enumerate(self._input_variables):
                value = "cluster%d"% (i + 1)
                if (j,i) in self._fuzzy_sets_to_drop.keys():
                    value = "cluster%d" % (self._fuzzy_sets_to_drop[(j, i)] + 1)
                pieces.append("(%s IS %s)" % (var, value)) 
                
            chunk = (" AND ".join(pieces))
            result.append( chunk )

        return result
        
    def create_rules(self):
        A = self._create_antecedents()
        # B = self._create_consequents()
        B = ["fun%d" % (i+1) for i in range(self._clusters)]
        result = ["IF %s THEN (OUTPUT IS %s)" % (a,b) for a,b in zip(A,B)]
        return result


if __name__ == '__main__':
    
    SC = SimpfulConverter(
        input_variables_names = ["pippo", "pluto"],
        consequents_matrix = [[1,2,3], 
                              [2,3,5]],
        fuzzy_sets = [
                ["gauss", [0,1]],
                ["sigmoid", [1,2]],
                ["gauss2", [0,1,2,3]],
                ["invgauss", [0,1]]
                ]
    )
    
    SC.save_code("TEST.py")
    SC.generate_object()
    print(FS._mfs['pippo'])
    a


import numpy as np
from scipy.optimize import curve_fit

def is_complete(G):
	nodelist = G.nodes
	H = G.subgraph(nodelist)
	n = len(nodelist)
	return H.size() == n*(n-1)/2

class AntecedentEstimator(object):
	def __init__(self, x_train, partition_matrix):
		self.xtrain = x_train 
		self.partition_matrix = partition_matrix
		self._info_for_simplification = None
		self._calculate_all_extreme_values()
						
		
	def determineMF(self, mf_shape='gauss', merge_threshold=1.):
		mf_list=[]

		# mf_list is structured as follows:
		# - an outer list of V variables
		# - each item of the outer list contains C fuzzy set, one for each cluster
		number_of_variables = self.xtrain.shape[1]
		for i in range(0, number_of_variables):
			xin = self.xtrain[:,i]
			for j in range(0, self.partition_matrix.shape[1]):
				mfin = self.partition_matrix[:,j]
				mf, xx = self.convexMF(xin=xin, mfin=mfin)
				prm = self.fitMF(x=xx, mf=mf, mf_shape=mf_shape)
				mf_list.append(prm) 

		self._check_similarities(mf_list, number_of_variables, threshold=merge_threshold)

		#print(self._info_for_simplification)

		return mf_list

	def is_subclique(self,G,nodelist):
		H = G.subgraph(nodelist)
		n = len(nodelist)
		return H.size() == n*(n-1)/2

	def _extreme_values_for_variable(self, v):
		return min(self.xtrain.T[v]), max(self.xtrain.T[v])

	def _calculate_all_extreme_values(self):
		num_variables = len(self.xtrain.T)
		self._extreme_values = [self._extreme_values_for_variable(v) for v in range(num_variables)]

	def _check_similarities(self, mf_list, number_of_variables,
			threshold=1., approx_points=100):

		number_of_clusters = len(mf_list)//number_of_variables
		
		from collections import defaultdict

		things_to_be_removed = defaultdict(list)


		""" 
			This function assesses the pair-wise similarities between 
			the clusters mapped on each variable.
			It returns a dictionary of this kind:
			variable -> list of similar couples for that variable + jaccard sim

		"""

		for v in range(number_of_variables):

			mi, ma = self._extreme_values_for_variable(v)
			points = np.linspace(mi, ma, approx_points)

			for c1 in range(number_of_clusters):
				for c2 in range(c1+1, number_of_clusters):

					index1 = v*number_of_clusters + c1
					index2 = v*number_of_clusters + c2
					funname1, params1 = mf_list[index1]
					funname2, params2 = mf_list[index2]

					if funname1== "gauss":
						
						from numpy import linspace, array

						first_cluster = array([self.gaussmf(x, params1[0], params1[1]) for x in points])
						second_cluster = array([self.gaussmf(x, params2[0], params2[1]) for x in points])

						intersection = sum([min(x,y) for x,y in zip(first_cluster, second_cluster)])
						union		= sum([max(x,y) for x,y in zip(first_cluster, second_cluster)])

						jaccardsim = (intersection/union)

						if jaccardsim>threshold:
							things_to_be_removed[v].append([c1,c2,jaccardsim])

							#print("%.2f is fine" % jaccardsim)

					else:
						raise Exception("Not implemented yet")

		#for k,v in things_to_be_removed.items():			print (k, v)
		#exit()
		"""
			This function must return a dictionary of items like:
			(variable, cluster) -> mapped_cluster
		"""

		self._info_for_simplification = {}
		for var_num, value in things_to_be_removed.items():

			subcomponents = self._create_graph(value)

			for subcomp in subcomponents:
				#print (is_complete(subcomp))

				if is_complete(subcomp):
					retained = list(subcomp.nodes())[0]
					#print ("retain: %d" % retained)
					for el in list(subcomp.nodes()):
						if el!=retained:
							self._info_for_simplification[(var_num, el)]  = retained					
					
		dropped_stuff = self.get_number_of_dropped_fuzzy_sets()
		print (" * %d antecedent clauses will be simplified using a threshold %.2f" % (dropped_stuff, threshold))
		self._info_for_simplification

	def get_number_of_dropped_fuzzy_sets(self):
		return len(self._info_for_simplification)

	def _create_graph(self, list_of_arcs):
		from networkx import Graph, connected_components
		G = Graph()
		nodelist = []
		for arc in list_of_arcs:
			G.add_edge(arc[0], arc[1])
			nodelist.append(arc[0])
			nodelist.append(arc[1])
		S = [G.subgraph(c).copy() for c in connected_components(G)]
		return S
			
	   
	def convexMF(self, xin, mfin, norm=1, nc=1000):
		
		# Calculates the convex membership function that envelopes a given set of
		# data points and their corresponding membership values. 
		
		# Input:
		# Xin: N x 1 input domain (column vector)
		# MFin: N x 1correspoding membership values 
		# nc: number of alpha cut values to consider (default=101)
		# norm: optional normalization flag (0: do not normalize, 1 : normalize, 
		# default=1)
		#
		# Output:
		# mf: membership values of convex function
		# x: output domain values	
		
		# Normalize the membership values (if requested)
		if norm == 1:
			mfin = np.divide(mfin, np.max(mfin))
		
		# Initialize auxilary variables
		acut = np.linspace(0,np.max(mfin),nc)
		mf= np.full(2*nc, np.nan)
		x=np.full(2*nc, np.nan)
		
		if np.any(mfin>0):
			x[0] = np.min(xin[mfin>0])
			x[nc]=np.max(xin[mfin>0])
			mf[0]=0
			mf[nc] = 0 
		
		# Determine the elements in the alpha cuts	
		for i in range(0,nc):
			if np.any(mfin>acut[i]):
				x[i]=np.min(xin[mfin>acut[i]])
				x[i+nc]=np.max(xin[mfin>acut[i]])
				mf[i]=acut[i]
				mf[i+nc]=acut[i]
				
		#Delete NaNs
		idx=np.isnan(x)
		x=x[idx==False]
		mf=mf[idx==False]  
		
		# Sort vectors based on membership value (descending order)
		indmf=mf.argsort(axis=0)
		indmf=np.flipud(indmf)
		mf=mf[indmf]
		x=x[indmf]
		
		# Find duplicate values for x and onlykeep the ones with the highest membership value
		_, ind = np.unique(x, return_index=True, return_inverse=False, return_counts=False, axis=None)
		mf=mf[ind]
		x=x[ind]
		
		# Sort vectors based on x value (ascending order)
		indx=x.argsort(axis=0)
		mf=mf[indx]
		x=x[indx]
		
		xval=np.linspace(np.min(x),np.max(x),nc)
		mf=np.interp(xval, x, mf, left=None, right=None, period=None)
		x=xval;
		return mf, x
	
	def fitMF(self,x,mf,mf_shape='gauss'):
		# Fits parametrized membership functions to a set of pointwise defined 
		# membership values.
		#
		# Input:
		# x:  N x 1 domain of input variable
		# mf: N x 1 membership values for input data x 
		# shape: Type of membership function to fit (possible values: 'gauss', 
		# 'gauss2' and 'sigmf')
		#
		# Output:
		# param: matrix of membership function parameters
	
		if mf_shape == 'gauss':
			# Determine initial parameters
			mu = sum(x * mf) / sum(mf)
			sig = np.mean(np.sqrt(-((x-mu)**2)/(2*np.log(mf))))
			
			# Fit parameters to the data using least squares
#			print('mu=', mu, 'sig=', sig)
			param, _ = curve_fit(self.gaussmf, x, mf, p0 = [mu, sig], bounds=((-np.inf, 0), (np.inf, np.inf)), maxfev = 10000)
	   
		elif mf_shape == 'gauss2':
			# Determine initial parameters
			mu1 = x[mf>=0.95][0]
			mu2 = x[mf>=0.95][-1]
			xmf =x[mf>=0.5]
			sig1 = (mu1 - xmf[0])/(np.sqrt(2*np.log(2)))
			sig2 = (xmf[-1]-mu2)/(np.sqrt(2*np.log(2)))
			if sig1==0.0:
				sig1=0.1
			if sig2==0.0:
				sig2=0.1
			
			# Fit parameters to the data using least squares
#			print('mu1',mu1,'sig1',sig1,'mu2',mu2,'sig2',sig2)
			param, _ = curve_fit(self.gauss2mf, x, mf, p0 = [mu1, sig1, mu2, sig2], bounds=((-np.inf, 0), (np.inf, np.inf)), maxfev=1000)
			
		elif mf_shape == 'sigmf':
			# Determine initial parameters
			if np.argmax(mf)-np.argmin(mf) > 0:		 # if sloping to the right
				c = x[mf>=0.5][0]
				s = 1
			elif np.argmax(mf)-np.argmin(mf) < 0:	   # if sloping to the left
				c = x[mf<=0.5][0]
				s = -1
			# Fit parameters of the function to the data using non-linear least squares		   
			param, _ = curve_fit(self.sigmf, x, mf, p0 = [c, s])
		
		return mf_shape, param
		
	def gaussmf(self,x, mu, sigma, a=1):
		# x:  (1D array)
		# mu: Center of the bell curve (float)
		# sigma: Width of the bell curve (float)
		# a: normalizes the bell curve, for normal fuzzy set a=1 (float) 
		return a * np.exp(-(x - mu)**2 / (2 * sigma**2))
	
	
	def gauss2mf(self,x, mu1, sigma1, mu2, sigma2):
		# x: Data
		# mu1: Center of the leftside bell curve
		# sigma1: Standard deviation that determines the width of the leftside bell curve
		# mu2: Center of the rightside bell curve
		# sigma2: Standard deviation that determines the width of the rightside bell curve
		y = np.ones(len(x))
		idx1 = x <= mu1
		idx2 = x > mu2
		y[idx1] = self.gaussmf(x[idx1], mu1, sigma1)
		y[idx2] = self.gaussmf(x[idx2], mu2, sigma2)
		return y
	
	def sigmf(self,x, c, s):
		# x: data
		# b: x where mf is 0.5
		# c: Controls 'width' of the sigmoidal region about `b` (magnitude); also
		#	which side of the function is open (sign). A positive value of `a`
		#	means the left side approaches 0.0 while the right side approaches 1,
		#	a negative value of `c` means the opposite.
		return 1. / (1. + np.exp(- s * (x - c)))


from collections import defaultdict
from math import sqrt
import numpy as np


class SugenoFISTester(object):
    """docstring for SugenoFISTester"""
    def __init__(self, model, test_data, golden_standard):
        super(SugenoFISTester, self).__init__()
        self._model_to_test = model
        self._data_to_test = test_data
        self._golden_standard = golden_standard

    def predict(self, variable_names, list_of_outputs=['OUTPUT']):
        result = []
        for sample in self._data_to_test:
            for i, variable in enumerate(variable_names):
                self._model_to_test.set_variable(variable, sample[i])
            result.append(self._model_to_test.Sugeno_inference().get('OUTPUT'))
        result = np.array(result) 
        error = self._golden_standard - result
        return result, error

    def calculate_RMSE(self, variable_names, list_of_outputs=['OUTPUT']):
        _, error=self.predict(variable_names, list_of_outputs)
        return sqrt(np.mean(np.square(error)))
    
    
    def calculate_MSE(self, variable_names, list_of_outputs=['OUTPUT']):
        _, error=self.predict(variable_names, list_of_outputs)
        return np.mean(np.square(error))   
    
    def calculate_MAE(self, variable_names, list_of_outputs=['OUTPUT']):
        _, error=self.predict(variable_names, list_of_outputs)
        return np.mean(np.abs(error))
    
    def calculate_MAPE(self, variable_names, list_of_outputs=['OUTPUT']):
        _, error=self.predict(variable_names, list_of_outputs)
        return np.mean(np.abs((error) / self._golden_standard)) * 100


import numpy as np

class DataSplitter(object):
    def __init__(self, dataX, dataY):
#        self.data=np.loadtxt(datapath,delimiter=',')
#        self.dataX=self.data[:,0:-1]
#        self.dataY=self.data[:,-1]
        self.dataX=dataX
        self.dataY=dataY
        
    def holdout(self,dataX,dataY, percentage_training=0.75,seed=None):
        universe=set(range(0,np.shape(dataX)[0]))
        trn=np.random.choice(dataX.shape[0], int(round(percentage_training*dataX.shape[0])), replace=False)
        tst=list(universe-set(trn))
    
        x_train=dataX[trn]
        x_test=dataX[tst]
        y_train=dataY[trn]
        y_test=dataY[tst]
        
        return x_train, y_train, x_test, y_test


    


from .simpfulfier import *

class SugenoFISBuilder(object):
    """docstring for SugenoFISBuilder"""
    def __init__(self, antecedent_sets, consequent_parameters, 
        variable_names, extreme_values=None, operators=None, save_simpful_code=True, fuzzy_sets_to_drop=None):
        
        super(SugenoFISBuilder, self).__init__()

        self._SC = SimpfulConverter(
            input_variables_names = variable_names,
            consequents_matrix = consequent_parameters,
            fuzzy_sets = antecedent_sets,
            operators = operators,
            extreme_values = extreme_values,
            fuzzy_sets_to_drop=fuzzy_sets_to_drop
            )
        
        if save_simpful_code==True:
            self._SC.save_code("Simpful_code.py")
        
        self._SC.generate_object()

        self.simpfulmodel = self._SC._fuzzyreasoner


from .BuildTakagiSugeno import *
from .Clustering import *
from .EstimateAntecendentSet import *
from .EstimateConsequentParameters import *
from .LoadData import *
from .simpfulfier import *
from .SimpfulModelBuilder import *
from .Splitter import *
from .Tester import *

import numpy as np

class pyFUME(object):
    def __init__(self, datapath, nr_clus, method='Takagi-Sugeno', variable_names=None, merge_threshold=1., **kwargs):
        self.datapath=datapath
        self.nr_clus=nr_clus
        self.method=method
        self.dropped_fuzzy_sets = 0
        #self.variable_names=variable_names

        if method=='Takagi-Sugeno' or method=='Sugeno':
            self.FIS = BuildTSFIS(self.datapath, self.nr_clus, variable_names, merge_threshold=merge_threshold, **kwargs)
            self.dropped_fuzzy_sets = self.FIS._antecedent_estimator.get_number_of_dropped_fuzzy_sets()
        else:
            raise Exception ("This modeling technique has not yet been implemented.")

    def get_model(self):
        if self.FIS.model is None:
            print ("ERROR: model was not created correctly, aborting.")
            exit(-1)
        else:
            return self.FIS.model

    def calculate_error(self, method="RMSE"):

        if method=="RMSE":
            return self._get_RMSE()
        elif method=="MAE":
            return self._get_MAE()
        elif method=="MAPE":
            return self._get_MAPE()
        elif method=="RMSE":
            return self._get_RMSE()
        else:
            # return self._get_MSE()
            raise Exception("Method '%s' not implemented yet" % (method))
            
    def predict_test_data(self):
        #get the prediction for the test data set
        test = SugenoFISTester(self.FIS.model, self.FIS.x_test, self.FIS.y_test)
        pred = test.predict(variable_names=self.FIS.variable_names)
        return pred
        
    def _get_RMSE(self):

        # Calculate the mean squared error of the model using the test data set
        test = SugenoFISTester(self.FIS.model, self.FIS.x_test, self.FIS.y_test)
        RMSE = test.calculate_RMSE(variable_names=self.FIS.variable_names)
        #RMSE = list(RMSE.values())
        #print('The RMSE of the fuzzy system is', RMSE)
        return RMSE
    
    def _get_MSE(self):

        # Calculate the mean squared error of the model using the test data set
        test = SugenoFISTester(self.FIS.model, self.FIS.x_test, self.FIS.y_test)
        MSE = test.calculate_MSE(variable_names=self.FIS.variable_names)
        #RMSE = list(RMSE.values())
        #print('The RMSE of the fuzzy system is', RMSE)
        return MSE
        
    
    def _get_MAE(self):

        # Calculate the mean absolute error of the model using the test data set
        test = SugenoFISTester(self.FIS.model, self.FIS.x_test, self.FIS.y_test)
        MAE = test.calculate_MAE(variable_names=self.FIS.variable_names)
        return MAE
    
    def _get_MAPE(self):

        # Calculate the mean absolute percentage error of the model using the test data set
        test = SugenoFISTester(self.FIS.model, self.FIS.x_test, self.FIS.y_test)
        MAPE = test.calculate_MAPE(variable_names=self.FIS.variable_names)
        return MAPE

    """
    def _get_RMSE(self):
        if self.FIS.error is None:
            print ("ERROR: RMSE was not calculated correctly, aborting.")
            exit(-1)
        else:
            return self.FIS.RMSE
    """ 

if __name__=='__main__':
    from numpy.random import seed
    seed(4)
   
    FIS = pyFUME(datapath='Concrete_data.csv', nr_clus=3, method='Takagi-Sugeno',
     merge_threshold=.8, operators=None)
    print ("The calculated error is:", FIS.calculate_error())

    FIS.get_model().produce_figure("bla.pdf")


from .SimpfulModelBuilder import SugenoFISBuilder
import numpy as np

class FireStrengthCalculator(object):
     def __init__(self, antecedent_parameters, variable_names, nr_clus, **kwargs):
           
    
        # Build a first-order Takagi-Sugeno model using Simpful using dummy consequent parameters
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
        simpbuilder = SugenoFISBuilder(
            self.antecedent_parameters, 
            np.tile(1, (self.nr_clus, len(self.variable_names)+1)), 
            self.variable_names, 
            extreme_values = self._antecedent_estimator._extreme_values,
            operators=kwargs["operators"], 
            save_simpful_code=False, 
            fuzzy_sets_to_drop=what_to_drop)

        self.dummymodel = simpbuilder.simpfulmodel
        
        # Calculate the firing strengths for each rule for each data point 
        firing_strengths=[]
        print(self.variable_names)
        for i in range(0,len(self.x_train)):
            for j in range (0,len(self.variable_names)):
                self.dummymodel.set_variable(self.variable_names[j], self.x_train[i,j])
            firing_strengths.append(self.dummymodel.get_firing_strenghts())
        self.firing_strengths=np.array(firing_strengths)
        


import setuptools

setuptools.setup(
    name = 'pyFUME', 
    version = '0.1.3',
    author = 'Caro Fuchs',
    author_email = 'c.e.m.fuchs@tue.nl',
    description = 'A Python package for fuzzy model estimation',
    long_description = open('README.md').read(),
    long_description_content_type =  'text/markdown',
    keywords = ['fuzzy logic', 'fuzzy inference systems', 'fuzzy model','data-driven', 'model estimation', 'machine learning'],
    url='https://github.com/CaroFuchs/pyFUME',
    packages=setuptools.find_packages(),
    license='LICENSE.txt',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[ 'numpy', 'simpful', 'fst-pso' ],
)


from .LoadData import DataLoader
from .Splitter import DataSplitter
from .SimpfulModelBuilder import SugenoFISBuilder
from .Clustering import Clusterer
from .EstimateAntecendentSet import AntecedentEstimator
from .EstimateConsequentParameters import ConsequentEstimator
from .Tester import SugenoFISTester
from .FeatureSelection import FeatureSelector
import numpy as np


class BuildTSFIS(object):
    def __init__(self, datapath, nr_clus, variable_names=None, merge_threshold=1.0, **kwargs):
        self.datapath = datapath
        self.nr_clus = nr_clus
        self.variable_names = variable_names
        self._antecedent_estimator = None
        
        # Load the data
        if 'normalize' not in kwargs.keys(): kwargs['normalize'] = False       
        dl=DataLoader(self.datapath,normalize=kwargs['normalize'])
        self.variable_names=dl.variable_names        

        # Split the data using the hold-out method in a training (default: 75%) 
        # and test set (default: 25%).
        ds = DataSplitter(dl.dataX,dl.dataY)
        self.x_train, self.y_train, self.x_test, self.y_test = ds.holdout(dl.dataX, dl.dataY)

        # Perform feature selection if requested
        if 'feature_selection' not in kwargs.keys(): kwargs['feature_selection'] = False
        if kwargs['feature_selection'] == True:
            fs=FeatureSelector(self.x_train, self.y_train, self.nr_clus, self.variable_names)
            
            self.selected_feature_indices, self.variable_names=fs.wrapper(feature_selection_stop=0.05)
            self.x_train = self.x_train[:, self.selected_feature_indices]
            self.x_test = self.x_test[:, self.selected_feature_indices]
            
        
        # Cluster the training data (in input-output space) using FCM
        cl = Clusterer(self.x_train, self.y_train, self.nr_clus)
        
        if 'cluster_method' not in kwargs.keys(): kwargs['cluster_method'] = 'fcm'
        
        if kwargs['cluster_method'] == 'fcm':
            if 'fcm_m' not in kwargs.keys(): kwargs['fcm_m'] = 2
            if 'fcm_max_iter' not in kwargs.keys(): kwargs['fcm_maxiter'] = 1000
            if 'fcm_error' not in kwargs.keys(): kwargs['fcm_error'] = 0.005
            self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fcm', fcm_m=kwargs['fcm_m'], 
                fcm_maxiter=kwargs['fcm_maxiter'], fcm_error=kwargs['fcm_error'])
        elif kwargs['cluster_method'] == 'fstpso':
            if 'fstpso_n_particles' not in kwargs.keys(): kwargs['fstpso_n_particles'] = None
            if 'fstpso_maxiter' not in kwargs.keys(): kwargs['fstpso_maxiter'] = 100
            if 'fstpso_path_fit_dump' not in kwargs.keys(): kwargs['fstpso_path_fit_dump'] = None
            if 'fstpso_path_sol_dump' not in kwargs.keys(): kwargs['fstpso_path_sol_dump'] = None
            self.cluster_centers, self.partition_matrix, _ = cl.cluster(cluster_method='fstpso', 
                fstpso_n_particles=kwargs['fstpso_n_particles'], fstpso_max_iter=kwargs['fstpso_max_iter'],
                fstpso_path_fit_dump=kwargs['fstpso_path_fit_dump'], fstpso_path_sol_dump=kwargs['fstpso_path_sol_dump'])
            
        
        # Estimate the membership funtions of the system (default shape: gauss)
        if 'mf_shape' not in kwargs.keys(): kwargs['mf_shape'] = 'gauss'       
        self._antecedent_estimator = AntecedentEstimator(self.x_train, self.partition_matrix)

        self.antecedent_parameters = self._antecedent_estimator.determineMF(mf_shape=kwargs['mf_shape'], merge_threshold=merge_threshold)
        what_to_drop = self._antecedent_estimator._info_for_simplification
        

        # Build a first-order Takagi-Sugeno model using Simpful using dummy consequent parameters
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
        simpbuilder = SugenoFISBuilder(
            self.antecedent_parameters, 
            np.tile(1, (self.nr_clus, len(self.variable_names)+1)), 
            self.variable_names, 
            extreme_values = self._antecedent_estimator._extreme_values,
            operators=kwargs["operators"], 
            save_simpful_code=False, 
            fuzzy_sets_to_drop=what_to_drop)

        self.dummymodel = simpbuilder.simpfulmodel
        
        # Calculate the firing strengths for each rule for each data point 
        firing_strengths=[]
        print(self.variable_names)
        for i in range(0,len(self.x_train)):
            for j in range (0,len(self.variable_names)):
                self.dummymodel.set_variable(self.variable_names[j], self.x_train[i,j])
            firing_strengths.append(self.dummymodel.get_firing_strengths())
        self.firing_strengths=np.array(firing_strengths)
        
        
        # Estimate the parameters of the consequent (default: global fitting)
        if 'global_fit' not in kwargs.keys(): kwargs['global_fit'] = False  
        ce = ConsequentEstimator(self.x_train, self.y_train, self.firing_strengths)
        self.consequent_parameters = ce.suglms(self.x_train, self.y_train, self.firing_strengths, 
                                               global_fit=kwargs['global_fit'])

        # Build a first-order Takagi-Sugeno model using Simpful
        if 'save_simpful_code' not in kwargs.keys(): kwargs['save_simpful_code'] = True
        if 'operators' not in kwargs.keys(): kwargs['operators'] = None
                   
        simpbuilder = SugenoFISBuilder(
            self.antecedent_parameters, 
            self.consequent_parameters, 
            self.variable_names, 
            extreme_values = self._antecedent_estimator._extreme_values,
            operators=kwargs["operators"], 
            save_simpful_code=kwargs['save_simpful_code'], 
            fuzzy_sets_to_drop=what_to_drop)

        self.model = simpbuilder.simpfulmodel
